/* (C) Copyright 1984,85,86,87 Walter L. Peacock   All Rights Reserved    */

/* b t r i s r t . c :                insert function for insert call     */
/*----------------------------------------------------------------------------
 * btrisrt() processes all INSERT and ISRTKY calls.  It invokes other
 * lower-level B+tree functions as needed to maintain the B+tree.
 *---------------------------------------------------------------------------*/

#include "cbtree.h"

btrisrt(fddta, fdidx, idxblk, hdrloc, node, rootptr, btcp)

BTC *btcp;
struct btidxblk *idxblk;
int fddta, fdidx, hdrloc;
long node, rootptr;
{
   extern long giveidxr(), pushdown(), poptop(), lseek();
   extern int strcmp(), strlen();
   extern char *strcpy(), *calloc();
   extern void free();
   extern long locklist[];
   extern int  lockptr;
   BTIDXBLK *nuidxblk, *idxp;
   long newnode, pos;
   int hloc, lk;
   char *newkey;
   uint splitpt;

   /**** NEWLOC ****/

   if (btcp->btoptype == NEWLOC)
   {
      if (btcp->btloc > 0)   /* we're located somewhere ! */
      {                            /* find continuation point   */
         while (btcp->btloc != idxblk->btpage[hdrloc].btptr)
         {
            if ((hdrloc = wrap_hdr(hdrloc+1, idxblk, fdidx, btcp))
                        == BTNOTFND)
               return (BTNOTFND);

            if (strcmp(btcp->btkey, idxblk->btpage[hdrloc].skeynme) > 0)
               return(BTNOTFND);
         }
      }    /* end if (btloc > 0) */

      /* set btloc to new location */
      btcp->btloc = idxblk->btpage[hdrloc].btptr = *btcp->btrecnum;

      /* and write it out */
      putidxr(fdidx, node, idxblk, btcp);

      return(BTCALLOK);
   }

   /**** INSERT and ISRTKY ****/

   if(idxblk->cellicnt < btcp->btcells)   /* we have room for new key */
   {

      insertky(idxblk, hdrloc, btcp->btkey, btcp->btkeylen);

      if (btcp->btoptype == INSERT) /* alloc new data rec. (locked) */
      {
         btcp->btloc = idxblk->btpage[hdrloc].btptr = btcp->btvarlen ?
               btgivvar(fddta, btcp->btdtalen) :
               giveidxr(fddta, btcp->btdtalen, 0, 1) ;
      }
      else                /* place rec pointer into node cell */
            idxblk->btpage[hdrloc].btptr = btcp->btloc;

      /* write the index record to disk  */
      putidxr(fdidx, node, idxblk, btcp);

      return(BTCALLOK);     /* return to calling pgm. */
   }

   /**** SPLIT THE NODE ****/

   /* this node is full: split node and insert key in proper node */

   /* make buffer for new node */
   if ((char *)(nuidxblk = (BTIDXBLK *)calloc(
         sizeof(BTIDXBLK) + btcp->btcells * sizeof(BTIDX), SZCHAR)) == NULL)
      ckerror(- CK_NOMEM, "35");

   splitpt = (idxblk->cellicnt + 1) >> 1;   /* (>> for LC bug) */
   if (idxblk->cellicnt & 1)    /* cell count is ODD */
      splitpt -= (hdrloc < splitpt);      /* adjust splitpoint */

   /* redistribute the cells between the two nodes */
   spltnode(idxblk, nuidxblk, splitpt);

   /*   now insert new key   */
   if (hdrloc < splitpt)      /* new key goes in THIS index block */
   {
      idxp = idxblk;
      hloc = hdrloc;
   }
   else     /* new key goes in THAT index block */
   {
      idxp = nuidxblk;
      hloc = hdrloc - (int)splitpt;
   }
   insertky(idxp, hloc, btcp->btkey, btcp->btkeylen);    /* add the key */
   if (btcp->btoptype == INSERT)   /* allocate new data record */
   {
      btcp->btloc = idxp->btpage[hloc].btptr = btcp->btvarlen ?
            btgivvar(fddta, btcp->btdtalen) :
            giveidxr(fddta, btcp->btdtalen, 0, 1) ;
   }
   else          /* place rec pointer into node cell */
      idxp->btpage[hloc].btptr = btcp->btloc;

   nuidxblk->blktype = idxblk->blktype;   /* set node type (blktype = 1) */

   /***** GET AN INDEX BLOCK FOR NEWLY CREATED NODE RECORD *****/

   /* working buffer for key */
   if ((newkey = calloc((uint)btcp->btkeylen + 1, SZCHAR)) == NULL)
      ckerror(- CK_NOMEM, "36");

   for(;;)  /* traverse all needed levels of the tree */
   {
      /* record location for new node */
      newnode = giveidxr(fdidx, btcp->btidxlen, 1, 0);
   
      locklist[lockptr++] = (newnode - 1L) * btcp->btidxlen;

      /* get copy of last key in node */
      strcpy(newkey, idxblk->btpage[idxblk->cellicnt - 1].skeynme);
   
      nuidxblk->bwdpage = node;
      nuidxblk->fwdpage = idxblk->fwdpage;
      idxblk->fwdpage = newnode;
   
      putidxr(fdidx, node, idxblk, btcp);    /* write the orginal node */
      putidxr(fdidx, newnode, nuidxblk, btcp);  /* write the new node */
   
      /* update bwdpage of fwdnode to point to newnode   */
      if (nuidxblk->fwdpage > 1)
      {
         /* locate the old right brother node so we can update the  */
		 /* backward pointer to point back to the newly inserted node. */
		 /* pos will point to beginning of node. */
         pos = (nuidxblk->fwdpage - 1L) * btcp->btidxlen;

         lk = 0;     /* lock the record we are about to modify */
         while (lockunlk(fdidx, pos, btcp->btidxlen, LOCKR) != 1)
            if (++lk > HDRTRYS){
               freekeys(nuidxblk);
               FREE(nuidxblk);
               FREE(newkey);
               return(BTLOCK); /* i.e. unable to lock linked list record */
			}

         /* now position to the backward pointer field of the node.  */
         if (lseek(fdidx, (long) (pos + (2 * sizeof(uword) + sizeof(long)) ),0) == -1L)
            ckerror(- CK_SEEK, "37");

         /* update the pointer */
         if (write(fdidx, (char *)&newnode, sizeof(long) ) == ERR)
            ckerror(- CK_WRIT, "38");

         lockunlk(fdidx, pos, btcp->btidxlen, UNLOCKR);
      }
    
      node = poptop();  /* get parent node */
    
      if (node == 0L){     /* then root node has just been split */

         /* MAKE A NEW ROOT NODE FOR THE TWO NEWLY SPLIT NODES */

         /* pick up LAST cells from both nodes */
         strcpy(idxblk->btpage[0].skeynme,
               idxblk->btpage[idxblk->cellicnt - 1].skeynme);
         idxblk->btpage[0].btptr = rootptr;  /* use old root pointer */
         strcpy(idxblk->btpage[1].skeynme,
               nuidxblk->btpage[nuidxblk->cellicnt - 1].skeynme);
         idxblk->btpage[1].btptr = newnode;

         idxblk->fwdpage = 0L;      /* initialize root node */
         idxblk->bwdpage = 0L;
         idxblk->cellicnt = 2;
         idxblk->blktype = 0;

         /* where to write new root pointer */
         rootptr = btcp->btroot = giveidxr(fdidx, btcp->btidxlen, 1, 0);
         putroot(fdidx, rootptr, btcp->findroot);  /* store new root addr */
    
         locklist[lockptr++] = (rootptr - 1L) * btcp->btidxlen;
         putidxr(fdidx, rootptr, idxblk, btcp);
   
         break;      /* all done */
      }

      /**** update PARENT node to reflect newly split children */
   
      getidxr(fdidx, node, idxblk, btcp->btidxlen, btcp->btkeylen);
   
      /* get btpage offset for position of key */
      hdrloc = binarysr(idxblk, newkey, strlen(newkey));
   
      /* now insert new key into previous node    */
      if (idxblk->cellicnt < btcp->btcells){   /* we have room for new key */
   
         insertky(idxblk, hdrloc, newkey, btcp->btkeylen);
         idxblk->btpage[hdrloc + 1].btptr = newnode;     /* set new node */
       
         /* write the index record to disk  */
         putidxr(fdidx, node, idxblk, btcp);

         break;   /* we done */
      }
   
      /**** PARENT NODE IS FULL: must be split ****/

      splitpt = (idxblk->cellicnt + 1) >> 1;   /* LC bug */
      if (idxblk->cellicnt & 1)    /* cell count is ODD */
         splitpt -= (hdrloc < splitpt);      /* adjust splitpoint */

      spltnode(idxblk, nuidxblk, splitpt);
   
      /* now insert new key */
      if(hdrloc < splitpt) {
         idxp = idxblk;
         hloc = hdrloc;
      }
      else {
         idxp = nuidxblk;
         hloc = hdrloc - (int)splitpt;
      }
      insertky(idxp, hloc, newkey, btcp->btkeylen);
      idxp->btpage[hloc + 1].btptr = newnode;
      nuidxblk->blktype = idxblk->blktype;   /* blktype = 0 */

   }  /* end for(;;): go back and write record */

   freekeys(nuidxblk);
   FREE(nuidxblk);
   FREE(newkey);
   return(BTCALLOK);           /* finally return to calling pgm. */
}

